﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.WebSockets;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class WebSocketSecureServer
    {
        private const int BUFFER_SIZE = 4096;

        /// <summary>
        /// OnConnect event
        /// </summary>
        public event Action<RemoteInfo> OnConnect;

        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<RemoteInfo> OnReceive;

        /// <summary>
        /// OnSend event
        /// </summary>
        public event Action<RemoteInfo> OnSend;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<RemoteInfo> OnDisconnect;

        /// <summary>
        /// The TLS version to be used by the server
        /// </summary>
        public SslProtocols TlsVersion { get; set; } = SslProtocols.Tls12;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        private readonly HttpListener _listener;
        private CancellationTokenSource _cancellationTokenSource;
        private Task _acceptTask;
        private readonly X509Certificate2 _serverCertificate;
        private readonly string _requiredPassword;
        private bool _passwordProtectionEnabled;

        private readonly Dictionary<string, WebSocket> _connectedClients = new Dictionary<string, WebSocket>();
        private readonly Dictionary<string, string> _clientNames = new Dictionary<string, string>();

        public CookieContainer Cookies { get; private set; } = new CookieContainer();
        public bool IsDebugHttpConnection { get; set; }

        /// <summary>
        /// Create secure WebSocket server with certificate support
        /// </summary>
        /// <param name="uriPrefixes">Array of URI prefixes to listen on, e.g., "https://localhost:8443/"</param>
        /// <param name="serverCertificate">The server certificate for HTTPS</param>
        /// <param name="requiredPassword">The password required from clients for the connection (optional)</param>
        public WebSocketSecureServer(List<string> uriPrefixes, X509Certificate2 serverCertificate = null, string requiredPassword = null)
        {
            _listener = new HttpListener();
            _serverCertificate = serverCertificate;
            _requiredPassword = requiredPassword;
            _passwordProtectionEnabled = !string.IsNullOrEmpty(requiredPassword);

            foreach (var uriPrefix in uriPrefixes)
            {
                if (uriPrefix.StartsWith("https"))
                {
                    _listener.Prefixes.Add(uriPrefix);
                }
                else if (uriPrefix.StartsWith("wss"))
                {
                    _listener.Prefixes.Add(uriPrefix.Replace("wss", "https"));
                }
                else
                {
                    throw new ArgumentException("Invalid URI prefix. Use 'https' or 'wss' prefixes.");
                }
            }
        }

        /// <summary>
        /// Start secure WebSocket server
        /// </summary>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        public Task StartAsync(CancellationToken cancellationToken = default)
        {
            _listener.Start();
            _cancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            _acceptTask = AcceptConnectionsAsync(_cancellationTokenSource.Token);
            return _acceptTask;
        }

        private async Task AcceptConnectionsAsync(CancellationToken cancellationToken)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var context = await _listener.GetContextAsync().ConfigureAwait(false);

                    if (IsDebugHttpConnection)
                    {
                        // Debug connection
                        DebugHttpsConnection(context);
                    }

                    if (context.Request.IsWebSocketRequest)
                    {
                        if (_passwordProtectionEnabled)
                        {
                            var password = context.Request.Headers["Password"];

                            if (password != _requiredPassword)
                            {
                                context.Response.StatusCode = 401; // Unauthorized
                                context.Response.Close();
                                continue;
                            }
                        }

                        // Manually set cookies in the HTTP response header
                        if (Cookies.Count > 0)
                        {
                            context.Response.Headers.Add("Set-Cookie", Cookies.GetCookieHeader(context.Request.Url));
                        }

                        string clientName = null;
                        foreach (var key in context.Request.Headers.AllKeys)
                        {
                            if (key == "ClientName")
                            {
                                clientName = context.Request.Headers["ClientName"];
                                break;
                            }
                        }

                        var webSocketContext = await context.AcceptWebSocketAsync(subProtocol: null).ConfigureAwait(false);
                        _ = HandleWebSocketConnectionAsync(webSocketContext.WebSocket, clientName, cancellationToken);
                    }
                    else
                    {
                        context.Response.StatusCode = 400;
                        context.Response.Close();
                    }
                }
                catch (Exception ex)
                {
                    OnError?.Invoke(ex, $"Exception: {ex.Message}");
                }
            }
        }


        private async Task HandleWebSocketConnectionAsync(WebSocket webSocket, string clientName, CancellationToken cancellationToken)
        {
            string clientId = Guid.NewGuid().ToString();
            _connectedClients.Add(clientId, webSocket);

            if (!string.IsNullOrEmpty(clientName))
            {
                _clientNames.TryAdd(clientId, clientName);
            }

            OnConnect?.Invoke(new RemoteInfo { IsWebSocket = true, ClientId = clientId, ClientName = clientName });

            byte[] buffer = new byte[BUFFER_SIZE];
            while (webSocket.State == WebSocketState.Open)
            {
                try
                {
                    var result = await webSocket.ReceiveAsync(new ArraySegment<byte>(buffer), cancellationToken).ConfigureAwait(false);
                    if (result.Count > 0)
                    {
                        byte[] data = new byte[result.Count];
                        Buffer.BlockCopy(buffer, 0, data, 0, result.Count);
                        OnReceive?.Invoke(new RemoteInfo { IsWebSocket = true, ClientId = clientId, ClientName = clientName, Data = data });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (Exception ex)
                {
                    OnError?.Invoke(ex, $"Exception: {ex.Message}");
                    break;
                }
            }

            OnDisconnect?.Invoke(new RemoteInfo { IsWebSocket = true, ClientId = clientId, ClientName = clientName });
            _connectedClients.Remove(clientId);
            _clientNames.Remove(clientId);
        }

        /// <summary>
        /// Stop secure WebSocket server
        /// </summary>
        /// <returns></returns>
        public async Task StopAsync()
        {
            _cancellationTokenSource.Cancel();
            _listener.Stop();
            if (_acceptTask != null)
            {
                try
                {
                    await _acceptTask.ConfigureAwait(false);
                }
                catch (AggregateException) { }
            }
        }

        /// <summary>
        /// Debug the HTTPS connection information
        /// </summary>
        /// <param name="context">HttpListenerContext representing the incoming connection</param>
        private void DebugHttpsConnection(HttpListenerContext context)
        {
            try
            {
                Console.WriteLine($"Incoming HTTPS Connection from: {context.Request.RemoteEndPoint}");

                // Output headers
                Console.WriteLine("Headers:");
                foreach (string key in context.Request.Headers.Keys)
                {
                    Console.WriteLine($"{key}: {context.Request.Headers[key]}");
                }

                // Check if the connection is WebSocket request
                if (context.Request.IsWebSocketRequest)
                {
                    Console.WriteLine("WebSocket Request detected.");

                    // Output WebSocket-specific information
                    if (_passwordProtectionEnabled)
                    {
                        var password = context.Request.Headers["Password"];
                        Console.WriteLine($"Password: {password}");
                    }

                    if (Cookies.Count > 0)
                    {
                        Console.WriteLine($"Cookies: {Cookies.GetCookieHeader(context.Request.Url)}");
                    }

                    foreach (var key in context.Request.Headers.AllKeys)
                    {
                        if (key == "ClientName")
                        {
                            var clientName = context.Request.Headers["ClientName"];
                            Console.WriteLine($"ClientName: {clientName}");
                            break;
                        }
                    }
                }
                else
                {
                    Console.WriteLine("Invalid WebSocket Request. Closing connection.");
                    context.Response.StatusCode = 400;
                    context.Response.Close();
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error while debugging HTTPS connection: {ex.Message}");
            }
        }
    }
}
